#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <math.h>

#define STR_HELPER(x) #x
#define STR(x) STR_HELPER(x)


//#define  ACCUM_SIZE  16       // Passed in as -D gcc parameter
//#define  CONSTANT_A  6        // Passed in as -D gcc parameter
//#define  CONSTANT_B  2        // Passed in as -D gcc parameter
//#define  CONSTANT_C  5        // Passed in as -D gcc parameter
//#define  CONSTANT_D  5        // Passed in as -D gcc parameter
//#define  pARITY               // Passed in as -D gcc parameter
//#define  SAMPLE_SIZE     16   // Passed in as -D gcc parameter


#if SAMPLE_SIZE > 17
	#error Sample size too big!  17 bits max.
#endif



#include "xoroshiro128lcg.c"




int  main(int argc, char* argv[])
{
	uint8_t  *pairs;
	uint32_t  pfreq[256];
	uint32_t  zfreq[256], freqi, ztally, next_prn, prev_prn;
#if ACCUM_SIZE > 16
	uint64_t  pairi;
#else
	uint32_t  pairi;
#endif
	uint64_t  total;
	FILE  *fh;
	uint8_t   fn[64];
	struct timespec  start, stop;
	double  duration;


	// Debugging the PRNG
//	for( irng = 0; irng < 80; irng++ )
//	{
//		printf( "%04x ", nextxoro() );
//	}
//	printf( "\n\n" );
//	exit(0);

	// initialize pair tally array
	pairs = (uint8_t *)calloc( 1UL<<(SAMPLE_SIZE*2), sizeof(uint8_t) );    // Zero'd memory allocation - big block

	if( pairs == NULL )
	{
		printf( "malloc error\n" );
		exit(1);
	}

	clock_gettime( CLOCK_MONOTONIC, &start );
	printf( "Started at %.3f   Malloc size = 0x%lx\n", start.tv_sec + start.tv_nsec/1.0e9, 1UL<<(SAMPLE_SIZE*2) );


	// initialize prev_prn
	prev_prn = nextxoro();

	// iterate full period of the PRNG and tally its output pair occurances
	// random ordered indexing - slowest memory writes (possible read-modify-write)
	pairi = (1UL<<(ACCUM_SIZE*2))-1;
	do {
		next_prn = nextxoro();
		pairs[prev_prn | (next_prn << SAMPLE_SIZE)]++;  // 8-bit limited tallies
		prev_prn = next_prn;

	} while( --pairi );


	clock_gettime( CLOCK_MONOTONIC, &stop );
	duration = (time_t)(stop.tv_sec - start.tv_sec) + (long)(stop.tv_nsec - start.tv_nsec)/1.0e9;
	printf( "[%d %d %d%s] paired at %.3f  -  duration = %.3f\n",
		CONSTANT_A, CONSTANT_B, CONSTANT_C, (CONSTANT_D>0?" "STR(CONSTANT_D):""), stop.tv_sec + stop.tv_nsec/1.0e9, duration );


	// initialize freq arrays
	memset( pfreq, 0, sizeof(pfreq) );
	memset( zfreq, 0, sizeof(zfreq) );

	// read pair array sequentially forward (fast) and increment freq values (fits in L1 caches easy)
	pairi = 0;
	ztally = 0;
	do {
		// pair freq values
		freqi = pairs[pairi++];
		pfreq[freqi]++;         // 8-bit limited indexing

		// zero-run freq values
		if( freqi == 0 )  ztally++;
		else if( ztally != 0 )
		{
			if( ztally > 255 )  ztally = 255;
			zfreq[ztally]++;
			zfreq[0]++;
			ztally = 0;
		}
	} while( pairi < (1UL<<(SAMPLE_SIZE*2))-1 );

	// pair freq values
	freqi = pairs[pairi];
	pfreq[freqi]++;

	// zero-run freq values
	if( freqi == 0 )  ztally++;

	if( ztally != 0 )
	{
		if( ztally > 255 )  ztally = 255;
		zfreq[ztally]++;
		zfreq[0]++;
	}


	clock_gettime( CLOCK_MONOTONIC, &stop );
	duration = (time_t)(stop.tv_sec - start.tv_sec) + (long)(stop.tv_nsec - start.tv_nsec)/1.0e9;
	printf( "[%d %d %d%s] finished at %.3f  -  total duration = %.3f\n",
		CONSTANT_A, CONSTANT_B, CONSTANT_C, (CONSTANT_D>0?" "STR(CONSTANT_D):""), stop.tv_sec + stop.tv_nsec/1.0e9, duration );


	#ifdef pARITY
		#define NAMESTR "-s%dp-a%db%dc%dd%dw%d"
		#define DESCSTR "Xoroshiro%d%sp [%d %d %d%s], sample width %d"
	#else
		#define NAMESTR "-s%d-a%db%dc%dd%dw%d"
		#define DESCSTR "Xoroshiro%d%s [%d %d %d%s], sample width %d"
	#endif
	// write freq array to file
	//   - four files, two binary and two text
	snprintf( fn, sizeof(fn), "pdist"NAMESTR".bin", ACCUM_SIZE, CONSTANT_A, CONSTANT_B, CONSTANT_C, CONSTANT_D, SAMPLE_SIZE );
	fh = fopen( fn, "w" );
	if( fh != NULL )
	{
		fwrite( pfreq, sizeof(pfreq), 1, fh );
		fclose( fh );
	}
	else	printf( "pbin file error\n" );


	// second binary file
	snprintf( fn, sizeof(fn), "zdist"NAMESTR".bin", ACCUM_SIZE, CONSTANT_A, CONSTANT_B, CONSTANT_C, CONSTANT_D, SAMPLE_SIZE );
	fh = fopen( fn, "w" );
	if( fh != NULL )
	{
		fwrite( zfreq, sizeof(zfreq), 1, fh );
		fclose( fh );
	}
	else	printf( "zbin file error\n" );


	snprintf( fn, sizeof(fn), "pdist"NAMESTR".txt", ACCUM_SIZE, CONSTANT_A, CONSTANT_B, CONSTANT_C, CONSTANT_D, SAMPLE_SIZE );
	fh = fopen( fn, "w" );
	if( fh != NULL )
	{
		fprintf( fh, "Frequency distribution of pairs for "DESCSTR"\n", ACCUM_SIZE*2, (CONSTANT_D>0?"++":"+"), CONSTANT_A, CONSTANT_B, CONSTANT_C, (CONSTANT_D>0?" "STR(CONSTANT_D):""), SAMPLE_SIZE );

		total = 0;
		for( freqi = 0; freqi < 256; freqi++ )
		{
			total += pfreq[freqi];
			fprintf( fh, "%3d  %u\n", freqi, pfreq[freqi] );
		}

		fprintf( fh, "Total = %lu or 0x%lx\n\n", total, total );
		fclose( fh );
	}
	else	printf( "ptxt file error\n" );


	snprintf( fn, sizeof(fn), "zdist"NAMESTR".txt", ACCUM_SIZE, CONSTANT_A, CONSTANT_B, CONSTANT_C, CONSTANT_D, SAMPLE_SIZE );
	fh = fopen( fn, "w" );
	if( fh != NULL )
	{
		fprintf( fh, "Frequency distribution of zero runs for "DESCSTR"\n", ACCUM_SIZE*2, (CONSTANT_D>0?"++":"+"), CONSTANT_A, CONSTANT_B, CONSTANT_C, (CONSTANT_D>0?" "STR(CONSTANT_D):""), SAMPLE_SIZE );

		total = 0;
		for( freqi = 0; freqi < 256; freqi++ )
		{
			total += zfreq[freqi];
			fprintf( fh, "%3d  %u\n", freqi, zfreq[freqi] );
		}

		fprintf( fh, "Total = %lu or 0x%lx\n\n", total, total );
		fclose( fh );
	}
	else	printf( "ztxt file error\n" );


	free( pairs );
	return 0;
}
