#!/bin/bash
accumsize=$1
algorithm=$2



case "$algorithm" in
"+")
	algorithm="+"     # single summing scrambler
	parity=""         # no parity in scrambler
    ;;
"+p")
	algorithm="+"     # single summing scrambler
	parity="-DpARITY" # scrambler parity at output bit0
    ;;
"++")
	algorithm="++"    # double summing scrambler
	parity=""         # no parity in scrambler
    ;;
"++p")
	algorithm="++"    # double summing scrambler
	parity="-DpARITY" # scrambler parity at output bit0
    ;;
*) # Default case
	echo Aborted!  Specify algorithm: +, ++, +p, ++p.
	echo
	exit
    ;;
esac

if [ $accumsize -lt 8 ] || [ $accumsize -gt 17 ]; then  # limited to 16 GB malloc()
	echo Aborted!  Specify size range 8-17.
	echo
	exit
fi


#enoughram=$((5 * 2**20))   # amount of spare ram, in KB, required before running next test case
enoughram=$(( 2**($accumsize*2-10) + 2**20 ))   # 16*2-10=22 => 4G+1G=5G  (-10 for 8-bit elements, -9 for 16-bit elements)

enginedir="tests/engine-xo-s${accumsize}"
testdir="${enginedir}/freq_dist"

if [ ! -f "${enginedir}/maxiter-xo-s${accumsize}.out" ]; then
	echo Aborted!  Test set not found.
	echo
	exit
fi


if [ ! -d "${testdir}" ]; then
	mkdir "${testdir}"
fi



function  test_case {

	RNGbin="case-s${accumsize}-a${a}b${b}c${c}d${d}"

	# Compile candidate test case
	gcc -o "${testdir}/${RNGbin}" "src/freq_dist.c" -O3 -DACCUM_SIZE=$accumsize -DCONSTANT_A=$a -DCONSTANT_B=$b -DCONSTANT_C=$c -DCONSTANT_D=$d -DSAMPLE_SIZE=$accumsize $parity

	if [ $? -ne 0 ]; then
		echo Distribution test case aborted!  Compile error.
		echo
		exit 1
	fi

	# Run test cases, concurrently
	pwdir=$( pwd )
	cd "${testdir}"
	nice -10 stdbuf -o L -e L "./${RNGbin}" &
	cd "${pwdir}"

	# Check for spare MIPS and spare RAM
	while true
	do
		sleep 1

#		spareram=$( awk '/MemAvailable/{print $2}' /proc/meminfo )
		read cpu spareram guest <<<"$( grep "MemAvailable:" /proc/meminfo )"

		read cpu user nice system idle iowait irq softirq steal guest <"/proc/stat"
		activeafter=$(( $user + $nice + $system + $softirq + $steal ))
		totalafter=$(( $activeafter + $idle + $iowait ))

		currload=$(( 101 * ($activeafter - $activebefore) / ($totalafter - $totalbefore) ))
		activebefore=$activeafter
		totalbefore=$totalafter

		cpuload=$(( $currload > $prevload ? $currload : $prevload ))  # max()
		prevload=$currload

		if [ $cpuload -lt 50 ] && [ $spareram -gt $enoughram ]; then  # Suited to dual threaded processor cores
			break
		fi
#		printf "cpuload=${currload}  \r" >&2

		sleep 0.5
	done

	rm "${testdir}/$RNGbin"	# Delete the binary, even though tests are still running
}



function  candidate_cases {
	printf "\n\n__________________________________________________________________________________________________________\n    BEGIN  ... at "
	date +"%F %T"

	if [ "$algorithm" == "+" ]; then # single summing scrambler
		string1="+${parity:2:1}"
	else # double summing scrambler
		string1="+${parity:2:1}+"
	fi
	printf "Single iterated Xoroshiro$(($accumsize*2))(${accumsize})${string1} random generator.\n\n"

	read cpu user nice system idle iowait irq softirq steal guest <"/proc/stat"
	activebefore=$(( $user + $nice + $system + $softirq + $steal ))
	totalbefore=$(( $activeafter + $idle + $iowait ))
	prevload=0

#a=3;b=2;c=13;d=6
#a=14;b=2;c=7;d=5
#test_case
#exit 0

	while IFS=", \t" read a b c line
	do
		if [ "$algorithm" == "+" ]; then # single summing scrambler
			printf "Pairs and Zero-run frequency distribution data for candidate [${a} ${b} ${c}]\n"
			d=0
			test_case

		else # Double summing scrambler

			for d in `seq 1 $(($accumsize-1))`
			do
				printf "Pairs and Zero-run frequency distribution data for candidate [${a} ${b} ${c} ${d}]\n"

				test_case
			done
		fi
	done < "${enginedir}/maxiter-xo-s${accumsize}.out"
}


candidate_cases |tee -a "${enginedir}/dist-xo-s${accumsize}.log"

